package History;
=head1 NAME
	History - Update history of contents for CGI.

	$Id: History.pm,v 1.2 2002/04/20 13:34:21 T.Shigetome Exp $

=head1 SYNOPSIS

	use Whatnew;
	use History;

	my($html_root);
	$html_root = '/home/foo/public_html';
	my(%config) = (
		'link'		=>	'rel',
		'html_root'	=>	"$html_root=http://www.any.com/~foo",
		'path'		=>	"$html_root",
		'template'	=>	"history=$html_root/work/histmpl.html;index=$html_root/work/idxtmpl.html",
		'records'	=>	"$html_root/work/whatnew.tmp",
		'index'		=>	"$html_root/work/history.idx=$html_root/history.html",
	);
	my($form) = new History;
	$form->ParseQuery();
	$form->SetParameter(\%config);
	$form->OutResponse();

=head1 DESCRIPTION

This is for CGI program, but may be used only by author.
This will create update history from temporary file of F<Whatnew.pm>.
I think the F<Whatnew.pm> has enough capability.
But the F<Whatnew.pm> forget expired old informations.
This is for keep that information to permanent file.
That permanent file is as HTML file.
Also, if you require, you can modify saved file.
This ability will give flexibility to your contents.

=head2 CGI ARGUMENT

This module will be used for CGI program. Like as... 

	http://www.any.com/~foo/cgi-bin/sample.cgi?config=test;func=history

or

	http://www.any.com/~foo/cgi-bin/sample.cgi?config=test;func=index

=over 4

=item func

There are two functions available.

=over 4

=item history

Append new update informations to history files.
If there is no history file, it will be created in specified path.
Update informations are formatted with template file.
Each history files can contain up to 10 entries.
When it reach to limit, next one will be created.

=item index

Create index list of update history files.
Will be re-indexed all files whenever execute this function.

=back

=item config

Value of `config' in QUERY STRING.

That value is used for hint of selecting from parameter set.
I hope to use with F<Params.pm>.

=back

=head2 METHOD

=over 4

=item new([$cgi_object])

Create new instance

	my($form) = new History;

or

	my($q) = new CGI;
	my($form) = new History($q);

=item ParseQuery

Parse QUERY_STRING

=item config

Return configuration name of `config' from CGI query string.
I think, that configuration name is useful with F<Params.pm>. ;) 

Like as... sample.cgi

	use Whatnew;
	use History;

	my($html_root) = '/home/foo/public_html';
	my($CONFIG_FILE) = "$html_root/work/sample.par";
	my($form) = new History;

	$form->ParseQuery();
	$main::configname = $form->config();
	$form->SetParameter($CONFIG_FILE);
	$form->OutResponse();

sample.par

	###########################################
	# History sample paramter file
	#
	#define	HTML_ROOT	/home/foo/public_html
	#define	WORKPATH %HTML_ROOT%/work
	#if $main::configname eq 'test'
	#	define	CONFIG_NAME	test
	#endif
	###########################################
	# common define
	%CONFIG_NAME%	link{
		rel;
	};
	%CONFIG_NAME%	html_root{
		%HTML_ROOT%=http://www.any.com/~foo
	};

	%CONFIG_NAME%	path{
		%HTML_ROOT%;
	};
	%CONFIG_NAME%	template{
		history=%WORKPATH%/histmpl.html;
		index=%WORKPATH%/idxtmpl.html;
	};
	###########################################
	test	records{
		%WORKPATH%/whatnew.tmp;
	};
	test	index{
		%WORKPATH%/history.idx=%HTML_ROOT%/history.html;
	};

=item SetParameter($rh_para)

Setup of parameter.
The $rh_para is reference of hash or a file name of parameter file.
If you specify a parameter file name, that parameter file will be parsed with F<Params.pm>.

	my($html_root);
	$html_root = '/home/foo/public_html';
	my(%config) = (
		'link'		=>	'rel',
		'html_root'	=>	"$html_root=http://www.any.com/~foo",
		'path'		=>	"$html_root",
		'template'	=>	"history=$html_root/work/histmpl.html;index=$html_root/work/idxtmpl.html",
		'records'	=>	"$html_root/work/whatnew.tmp",
		'index'		=>	"$html_root/work/history.idx=$html_root/history.html",
	);
	my($form) = new History;
	$form->ParseQuery();
	$form->SetParameter(\%config);

or

	my($CONFIG_FILE) = "$html_root/work/sample.par";
	my($form) = new History;

	$form->ParseQuery();
	$main::configname = $form->config();
	$form->SetParameter($CONFIG_FILE);


=over 4

=item link

Specify mode of path relative or absolute.
The `rel' is relative URL, and the `abs' is absolute URL.

=item html_root

Specify file path connection with URL path.

=item path

Specify path for files that generated by this.

=item template

Specify template files for history files and index files.
Each template file names join keywords by `='.
Like as...

	`history=histmpl.html;index=idxtmpl.html'

The `histmpl.html' is for history file,
 and the `idxtmpl.html' is for index file.

You can easily customize template.
Sample is following.

	<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01//EN">
	<HTML>
	<HEAD>
	<META HTTP-EQUIV="Content-Type" CONTENT="text/html; charset=Shift_JIS">
	<TITLE>Update history index</TITLE>
	<!--
		Update history sample template
	-->
	</HEAD>
	<BODY>
	<H1>Update history</H1>
	<HR>
	<TABLE summary="Update history">
	<!--
		Following comment statement will be replaced to new documents information.
		You can customize that.
		These keywords will be handled about associated document.
		Like as...

		%mtime%	: Modified time
		%URL%	: URL
		%title%	: Title in head block of that document
		%x-meta-updateinfo%
			: Update information
			  It's described in head block of that document.
			  If you want to use this field, you must append it along specified format when you update document.
	-->
	<!-- NEW items 0 0
	<TR>
	<TD VALIGN="top">%mtime%</TD><TD VALIGN="top">:</TD><TD><P><A HREF="%URL%">%title%</A></P><PRE>%x-meta-updateinfo%</PRE></TD>
	</TR>
	NEW items to above -->
	</TABLE>
	<HR>
	</BODY>
	</HTML>


=item records

Specify update information records that is created by F<Whatnew.pm>.
If you want to two or more, you can describe separated each one by `;'.
Do NOT insert space character between each define.

You can use URLs, if you use merge function in the Whatnew.
For example...

	http://www.another.com/~foo/cgi-bin/whatnew.cgi&config=another;func=get

=item index

Specify temporary file of index and final formatted index file.

=back

=item OutResponse

Output result as response along specified function by`C<func>'.

=item debug

Toggle each debug condition.
Default is debug off.

=over 4

=item configpara

Debug of parameter.

=item template

Debug of parsing template.

=item history

Debug of history information append.

=item index

Debug of generating index.

=item remote

Debug of getting records via HTTP.

=item records

Debug of record

=back

=back

=head1 SEE ALSO

F<Whatnew.pm>, F<Params.pm>, F<Timestr.pm>, F<CGI.pm>, F<URI.pm>

=head1 AUTHOR

Tetsuya Shigetome, E<lt>t_shigetome@muf.biglobe.ne.jpE<gt>.

=head1 HISTORY

=over 4

=item $Revision: 1.2 $

$Log: History.pm,v $
Revision 1.2  2002/04/20 13:34:21  T.Shigetome
Fixed bugs

o Sort order is not numerically.

o Foever loop in UpdateHistory.

Compliant

Whatnew.pm and Params.pm


=item Revision: 1.1

Log: History.pm,v 
Revision 1.1  2001/03/31 13:42:57  shige

First release of History.pm


=back

=cut
$History::VERSION = sprintf("%d.%02d", q$Revision: 1.2 $ =~ /(\d+)\.(\d+)/);

use strict;
use Carp;
use CGI;
use File::Basename qw(fileparse);
use URI;
use Params;
use Timestr;


######################
# Constructer
sub new
{
my($pkg,$cgi_query) = @_;
my($rh_this);
	$rh_this =  bless {
#>>>DEBUG_CODE
		'_debug'			=>	{
			'configpara'	=>	undef,
			'template'		=>	undef,
			'history'		=>	undef,
			'index'			=>	undef,
			'remote'		=>	undef,
			'records'		=>	undef,
		},#  'configpara '
#>>>DEBUG_CODE
		'config'		=>	undef,
		'base'			=>	undef,
# Following value may be over load by user
		'limit'			=>	10, # limit of history record count in each file
		'path'			=>	undef, # History folder info
		'filename'		=>	{
			'format'		=> 'his%05d.html',			# format string for sprintf
			'regular_expr'	=> 'his([0-9]{5})\.html',	# match expression
			'temp'			=>	"$$.tmp",				# temporary filename
			'name'			=>	undef,					# current filename
		},
		'link'			=>	'abs',
# untile
#
		'_BoundaryString'	=>	undef,#{},
		'_UpdateInfo'		=>	{
			'buffer'		=>	[],
			'URI'			=>	[],
			'fieldlist'		=>	[],
			}, # [] file name or url
		'_TemplateFile'	=>	{
			'history'		=>	undef,	# used for history file
			'index'			=>	undef,	# used for index of history
			'buffer'		=>	undef,	# cache of hitory template
		},
		'_HistoryFile'		=>	{
			'buffer'		=>	undef,
			'_ReplaceKeys'	=>	{
									'<!-- NEW items\s+([0-9]+)\s+([0-9]+)\s+(.+)NEW items to above -->\s*'	=>	\&NewItemsExtender,
			},
			'relative'		=>	1,
		},
		'_HistoryIndex'		=>	{
			'index'			=>	{},		# index record file
			'buffer'		=>	{},		# index record buffer
			'outbuffer'		=>	{},	# output html data
			'fieldlist'		=>	['mtime','filename','entry'],
			'array'			=>	{},
			'_ReplaceKeys'	=>	{
									'<!-- History files\s+(.+)History files to above -->\s*'	=>	\&HistoryIndexExtender,
			},
			'File2URL'		=>	\&File2RelURL,
		},
		'_HtmlRoot'			=>	{
		},
		'_ParameterEvaluater'	=>	undef,#{},	# It' described later
		'_CgiQuery'			=>	undef,
	}, $pkg;
	$rh_this->{'_CgiQuery'} = $cgi_query if defined($cgi_query);
#######################
# Defined CGI argument evaluation subroutine
#
	$rh_this->{'_ParameterEvaluater'} = {
#//////////////////////////////////////////
		'config' => sub {
		my($pkg,$string) = @_;
			$pkg->{'config'} = $string;
		},
#//////////////////////////////////////////
		'func' => sub {
		my($pkg,$string) = @_;
		my(%ResponseMethod) = (
							'history'	=>	\&OutResultOfHistoryUpdate,
							'index'		=>	\&OutResultOfCreateHistoryIndex,
							);
			# select Response method
			die "Unknown funtion : $string" unless ( defined($ResponseMethod{$string}) );
			$pkg->{'_Method'}{'_Response'} = $ResponseMethod{$string};
		},
	};

	return $rh_this;
}
#######################
# Method autoload
sub AUTOLOAD
{
my($pkg) = shift;
my($type) = ref($pkg) || croak "$pkg is not an object  on $History::AUTOLOAD";
my($name) = $History::AUTOLOAD;
	$name =~ s/.*://;
	unless( exists($pkg->{$name}) )
	{
#>>>DEBUG_CODE
		unless( $name eq 'debug' )
		{
#>>>DEBUG_CODE
			croak "Can't access `$name' field in object of class $type";
#>>>DEBUG_CODE
		}
		else
		{
			if( @_ )
			{
				foreach( @_ )
				{
					if( defined($pkg->{'_debug'}{$_}) )
					{
						undef $pkg->{'_debug'}{$_};
					}
					else
					{
						$pkg->{'_debug'}{$_} = 1;
					}
				}
			}
		}
		return wantarray ? keys(%{$pkg->{'_debug'}}) : $pkg->{'_debug'};
#>>>DEBUG_CODE
	}
	return $pkg->{$name} = shift if( @_ );
	return $pkg->{$name};
}
#	It's a scapegoat for AUTOLOAD.
sub	DESTROY{}
#######################
# Parse Query string
sub ParseQuery
{
my($pkg) = shift;
local($_);
my($key,$func);

	unless( defined($pkg->{'_CgiQuery'}) )
	{
		$pkg->{'_CgiQuery'} = new CGI;
	}
#>>>DEBUG_CODE
	else
	{
		warn if defined($pkg->{'_debug'}{'configpara'});
	}
#>>>DEBUG_CODE
	foreach $key ( ($pkg->{'_CgiQuery'}->param) )
	{
		$value = $pkg->{'_CgiQuery'}->param($key);
#>>>DEBUG_CODE
		warn "$key=$value" if defined($pkg->{'_debug'}{'configpara'});
#>>>DEBUG_CODE
		if( defined($pkg->{'_ParameterEvaluater'}{$key}) )
		{
			&{$pkg->{'_ParameterEvaluater'}{$key}}($pkg,$value);
		}
		else
		{# It may be extended parameter. But do NOT override.
			$pkg->{$key} = $value unless exists($pkg->{$key});
		}
	}
}
#######################
# Set parameter
sub SetParameter
{
my($pkg) = shift @_;
my($rh_para) = @_;
my(%para_parser) = (
	'path'		=>	[\&Set2ScalarRef,	\$pkg->{'path'},					qq/Unknown "path" argument type /],
	'records'	=>	[\&Set2ArrayRef,	\$pkg->{'_UpdateInfo'}{'URI'},		qq/Unknown "records" argument type /],
	'template'	=>	[\&Set2HashRef,		\$pkg->{'_TemplateFile'},			qq/Unknown "template" argument type /],
	'index'		=>	[\&Set2HashRef,		\$pkg->{'_HistoryIndex'}{'index'},	qq/Unknown "index" argument type /],
	'html_root'	=>	[\&Set2HashRef,		\$pkg->{'_HtmlRoot'},				qq/Unknown "html_root" argument type /],
	);
	unless( ref($rh_para) )
	{# It's specified paramter file
		my($para) = new Params($rh_para);
		my($configname) = $pkg->config();
		warn "Handled as parameter file $rh_para." if defined($pkg->{'_debug'}{'configpara'});
		croak "Do not specify configuration name in CGI query string" unless defined($configname);
		$rh_para = $para->GetData($configname);
	}
	foreach( (keys(%para_parser)) )
	{
		if( defined( $rh_para->{$_} ) )
		{
#>>>DEBUG_CODE
			warn "Parse parameter : $_ => ${$rh_para}{$_}" if defined($pkg->{'_debug'}{'configpara'});
#>>>DEBUG_CODE
			die "$para_parser{$_}[2] $pkg" unless &{$para_parser{$_}[0]}($para_parser{$_}[1],$rh_para->{$_});
		}
#>>>DEBUG_CODE
		else
		{
			warn "Undefined parameter : $_" if defined($pkg->{'_debug'}{'configpara'});
		}
#>>>DEBUG_CODE
	}
	if( defined($rh_para->{'link'}) )
	{
#>>>DEBUG_CODE
		warn "Parse parameter : link => ${$rh_para}{'link'}" if defined($pkg->{'_debug'}{'configpara'});
#>>>DEBUG_CODE
		$pkg->SetURIBase($rh_para->{'link'});
	}
	{# Get Whatnew variable
		croak "Undefined fieldlist in Whatnew that may be too old" unless defined(@Whatnew::fieldlist);
		$pkg->{'_UpdateInfo'}{'fieldlist'} = \@Whatnew::fieldlist;
		croak "Undefined BoundaryString in Whatnew that may be too old" unless defined(%Whatnew::BoundaryString);
		$pkg->{'_BoundaryString'} = \%Whatnew::BoundaryString;
	}
}
#######################
# Output Response
sub OutResponse
{
my($pkg) = shift;
	die "Unknown method of output response : $pkg" unless defined($pkg->{'_Method'}{'_Response'});
	&{$pkg->{'_Method'}{'_Response'}}($pkg);
}
#######################
# Output result of history update
sub OutResultOfHistoryUpdate
{
my($pkg) = shift;
my($updatefiles);
my($record_path,$indexpage_path);
my($record_temp,$indexpage_temp);
	$pkg->GetUpdateInformation();
	$updatefiles = join("\n",$pkg->UpdateHistory());

print <<"__HISTORY_UPDATE_RESULT__";
Content-type: text/plain

Finish history update !!

At following files

$updatefiles

__HISTORY_UPDATE_RESULT__
}
#######################
# Output result of create history index
sub OutResultOfCreateHistoryIndex
{
my($pkg,$rh_index) = shift;
my($record_path,$indexpage_path);
my($record_temp,$indexpage_temp);
my($index_filename,$html_filename);
	$rh_index = $pkg->{'_HistoryIndex'} unless defined($rh_index);

	$pkg->UpdateIndexRecord();
	$pkg->UpdateHistoryIndex();

	while( ($index_filename,$html_filename) = each(%{$rh_index->{'index'}})  )
	{
		# output to temporary
		($record_temp,$record_path) = fileparse($index_filename);
		($indexpage_temp,$indexpage_path) = fileparse($html_filename);
		$record_temp = $record_path.$record_temp.$$;
		$indexpage_temp = $indexpage_path.$indexpage_temp.$$;

		open(TEMP,"> $record_temp") || die "Can NOT create index record temporary $record_temp : $!";
		print TEMP $rh_index->{'buffer'}{$index_filename};
		close(TEMP);

#>>>DEBUG_CODE
		warn $indexpage_temp if defined($pkg->{'_debug'}{'index'});
#>>>DEBUG_CODE
		open(TEMP,"> $indexpage_temp") || die "Can NOT create index html temporary $indexpage_temp : $!";
		print TEMP $rh_index->{'outbuffer'}{$index_filename};
		close(TEMP);

		# rename
		rename($record_temp,$index_filename);
		rename($indexpage_temp,$html_filename);
	}

print <<"__HISTORY_UPDATE_RESULT__";
Content-type: text/plain

Finish history index update !!

__HISTORY_UPDATE_RESULT__
}
################################
# Load history file list
sub GetNewestHistory
{
my($pkg,$rh_filename,$path) = @_;
my(@work);
	$pkg->{'filename'}{'name'} = $pkg->GetHistoryFiles($rh_filename,$path);
	if( defined($pkg->{'filename'}{'name'}) )
	{# Any history exist
		$pkg->{'_HistoryFile'}{'buffer'} = '';
		$pkg->LoadFile("$path/$rh_filename->{'name'}",\$pkg->{'_HistoryFile'}{'buffer'}) || die "Can NOT open history file $path/$rh_filename->{'name'} : $!";
#>>>DEBUG_CODE
		warn "$path/$rh_filename->{'name'}" if defined($pkg->{'_debug'}{'template'});
#>>>DEBUG_CODE
	}
	else
	{# Copy template
		$pkg->LoadFile($pkg->{'_TemplateFile'}{'history'},\$pkg->{'_TemplateFile'}{'buffer'}) unless defined($pkg->{'_TemplateFile'}{'buffer'});
		$pkg->{'_HistoryFile'}{'buffer'} = $pkg->{'_TemplateFile'}{'buffer'};
		$rh_filename->{'name'} = sprintf($rh_filename->{'format'},1);
#>>>DEBUG_CODE
		warn $pkg->{'_TemplateFile'}{'history'} if defined($pkg->{'_debug'}{'template'});
		warn $rh_filename->{'name'} if defined($pkg->{'_debug'}{'template'});
#>>>DEBUG_CODE
	}
}
################################
# Load history file list
sub GetHistoryFiles
{
my($pkg,$rh_filename,$path) = @_;
my(@work);

	$rh_filename = $pkg->{'filename'} unless defined($rh_filename);
	$path = $pkg->{'path'} unless defined($path);

	die 'Do NOT specify history path' unless( defined($path) );
	opendir(DIR,$path) || die "Can't open history path $path: $!";
	@work = reverse sort (grep (/$rh_filename->{'regular_expr'}/ , readdir(DIR)));
	closedir(DIR);

	return wantarray ? @work : $work[0];
}
################################
# Load history index record
sub UpdateIndexRecord
{
my($pkg,$rh_index) = @_;
my($filename,$ra_index);
my(%hashed_index);
my($i);
my( @work );
my($history);
my(%fieldno);
my($history_filename,@historyfiles);
my(@record);
my(@history_files);

	$rh_index = $pkg->{'_HistoryIndex'} unless defined($rh_index);

	$i = 0;
	foreach( @{$pkg->{'_HistoryIndex'}{'fieldlist'}} )
	{
		$fieldno{$_} = $i;
		++$i;
	}

	foreach $filename ( keys(%{$rh_index->{'index'}}))
	{
		$rh_index->{'buffer'}{$filename} = '';
		$pkg->LoadFile($filename,\$rh_index->{'buffer'}{$filename});

		$rh_index->{'array'}{$filename} = [];
		$ra_index = $rh_index->{'array'}{$filename};

		{# update last record
			@{$ra_index} = sort {my($a_time,$b_time);$a =~ m/^\d+/; $a_time = $&;$b =~ m/^\d+/; $b_time = $&; $a_time <=> $b_time;} split(/$pkg->{'_BoundaryString'}{'record'}\s*/so,$rh_index->{'buffer'}{$filename} );
			if( @{$ra_index} > 0 )
			{
				@record = split(/$pkg->{'_BoundaryString'}{'field'}/so,$ra_index->[-1] );
				$history_filename = $record[$fieldno{'filename'}];

				$history = '';
				if( $pkg->LoadFile($history_filename,\$history) )
				{
					if( $history =~ m/<!-- NEW items\s+([0-9]+)\s+([0-9]+)\s+(.+)-->\s*/s )
					{
						$record[$fieldno{'mtime'}] = $1;
						$record[$fieldno{'entry'}] = $2;
						$ra_index->[-1] = join($pkg->{'_BoundaryString'}{'field'},@record);
					}
				}
				else
				{
					warn "Can NOT find $history_filename : $!";
				}
			}
		}
		# append new record
		foreach( @{$ra_index} )
		{
			$history_filename = (split(/$pkg->{'_BoundaryString'}{'field'}/so,$_ ))[$fieldno{'filename'}];
#>>>DEBUG_CODE
			warn $history_filename if defined($pkg->{'_debug'}{'index'});
#>>>DEBUG_CODE
			$hashed_index{$history_filename} = $_;
			++$i;
		}
		@historyfiles = $pkg->GetHistoryFiles();
		foreach( @historyfiles )
		{
			next if defined($hashed_index{"$pkg->{'path'}/$_"});
			$history = '';
			if( $pkg->LoadFile("$pkg->{'path'}/$_",\$history) )
			{
#>>>DEBUG_CODE
				warn "$pkg->{'path'}/$_" if defined($pkg->{'_debug'}{'index'});
#>>>DEBUG_CODE
				if( $history =~ m/<!-- NEW items\s+([0-9]+)\s+([0-9]+)\s+(.+)-->\s*/s )
				{
#>>>DEBUG_CODE
					if( defined($pkg->{'_debug'}{'index'}) )
					{
						warn "$pkg->{'path'}/$_" ;
						warn "mtime=$fieldno{'mtime'} entry=$fieldno{'entry'} filename=$fieldno{'filename'}";
						warn "$1 $2";
					}
#>>>DEBUG_CODE
					$record[$fieldno{'mtime'}] = $1;
					$record[$fieldno{'entry'}] = $2;
					$record[$fieldno{'filename'}] = "$pkg->{'path'}/$_";
#>>>DEBUG_CODE
					warn join($pkg->{'_BoundaryString'}{'field'},@record) if defined($pkg->{'_debug'}{'index'});
#>>>DEBUG_CODE
					push(@{$ra_index},join($pkg->{'_BoundaryString'}{'field'},@record));
				}
			}
#>>>DEBUG_CODE
			else
			{
				warn "Unmatch history :$pkg->{'path'}/$_" if defined($pkg->{'_debug'}{'index'});
			}
#>>>DEBUG_CODE
		}
		$rh_index->{'buffer'}{$filename} = join("$pkg->{'_BoundaryString'}{'record'}\n", @{$ra_index});
	}
	# return modify count
}
################################
# Load history index record
sub UpdateHistoryIndex
{
my($pkg,$rh_index) = @_;
my($filename,$ra_buffer);

	$rh_index = $pkg->{'_HistoryIndex'} unless defined($rh_index);

	while( ($filename,$ra_buffer) = each(%{$rh_index->{'array'}} ))
	{
		$rh_index->{'outbuffer'}{$filename} = '';

		if( $pkg->LoadFile($pkg->{'_TemplateFile'}{'index'},\$rh_index->{'outbuffer'}{$filename}) )
		{# Extend replace key string in history file
			warn $rh_index->{'outbuffer'}{$filename} if defined($pkg->{'_debug'}{'template'});
			foreach( keys(%{$rh_index->{'_ReplaceKeys'}}) )
			{
				my($work);
				warn "Can NOT find replacekey in template : $_"
					unless $rh_index->{'outbuffer'}{$filename} =~ s/$_/&{$rh_index->{'_ReplaceKeys'}{$_}}($pkg,$rh_index->{'index'}{$filename},$ra_buffer,$1)/sge;
			}
		}
#>>>DEBUG_CODE
		else
		{
			warn "Can NOT open template file for  history index $pkg->{'_TemplateFile'}{'index'} : $!";
		}
#>>>DEBUG_CODE
	}
}
################################
# Load file
sub LoadFile
{
my($pkg,$filename,$rs_buffer) = @_;
my($result);
local($_);

	$result = open(LOADFILE, "< $filename");
	if( $result )
	{
		while( <LOADFILE> )
		{# Load template
			$$rs_buffer .= $_;
		}
		close(LOADFILE);
	}
	return $result;
}
################################
# Get update informations
sub GetUpdateInformation
{
my($pkg,$rh_updateinfo) = @_;
my($buffer) = '';
#>>>NO_REMOTE
my($ua);
my($header);
my($request,$response);

	$rh_updateinfo = $pkg->{'_UpdateInfo'} unless defined($rh_updateinfo);

#>>>NO_REMOTE
	foreach( @{$rh_updateinfo->{'URI'}} )
	{
		next if /^\s*$/;
#>>>NO_REMOTE
		if( m|^\w+://|  )# include any schema
		{
			unless( defined($header) )
			{
				$ua = LWP::UserAgent->new;
				$header = new HTTP::Headers();
				$header->referer($ENV{'HTTP_REFERER'}) if defined($ENV{'HTTP_REFERER'});
				$header->referer($ENV{'HTTP_USER_AGENT'}) if defined($ENV{'HTTP_USER_AGENT'});
#>>>DEBUG_CODE
				warn $header->as_string() if defined($pkg->{'_debug'}{'remote'});
#>>>DEBUG_CODE
			}
#>>>DEBUG_CODE
			warn "Remote URL : $_" if defined($pkg->{'_debug'}{'remote'});
#>>>DEBUG_CODE
			$request = HTTP::Request->new('GET', $_, $header);

			$response = $ua->request($request);
			if( $response->is_success )
			{
#>>>DEBUG_CODE
				warn $response->{'_content'} if defined($pkg->{'_debug'}{'remote'});
#>>>DEBUG_CODE
				$buffer .= $response->{'_content'};
			}
#>>>DEBUG_CODE
			else
			{
				warn $response->status_line." : $_" if defined($pkg->{'_debug'}{'remote'});
			}
#>>>DEBUG_CODE

		}
		else
#>>>NO_REMOTE
		{
			unless( $pkg->LoadFile($_,\$buffer) )
			{
				warn "Can NOT open update information file $_ : $!";
			}
#>>>DEBUG_CODE
			warn $_ if defined($pkg->{'_debug'}{'records'});
#>>>DEBUG_CODE
		}
	}
#>>>DEBUG_CODE
	warn $buffer if defined($pkg->{'_debug'}{'records'});
	warn if defined($pkg->{'_debug'}{'records'});
#>>>DEBUG_CODE
	@{$rh_updateinfo->{'buffer'}} = split(/$pkg->{'_BoundaryString'}{'record'}\n*/,$buffer);
#>>>DEBUG_CODE
	warn $pkg->{'_BoundaryString'}{'record'} if defined($pkg->{'_debug'}{'records'});
	warn scalar(@{$rh_updateinfo->{'buffer'}}) if defined($pkg->{'_debug'}{'records'});
#>>>DEBUG_CODE
}
################################
# Update history
sub UpdateHistory
{
my($pkg,$ra_buffer,$rh_filename,$path) = @_;
my($tempfile);
my(@updatedfiles) = ();
my($infinity_check,$dirty) = (0,0);

	$ra_buffer = $pkg->{'_UpdateInfo'}{'buffer'} unless defined($ra_buffer);
	$rh_filename = $pkg->{'filename'} unless defined($rh_filename);
	$path = $pkg->{'path'}  unless defined($path);
	$tempfile = "$path/$pkg->{'filename'}{'temp'}";

	$pkg->GetNewestHistory($rh_filename,$path) unless defined($pkg->{'_HistoryFile'}{'buffer'});

	while( @{$ra_buffer} > 0 && $infinity_check != scalar @{$ra_buffer} )
	{
		$infinity_check = scalar @{$ra_buffer};
		# Extend replace key string in history file
		foreach( keys(%{$pkg->{'_HistoryFile'}{'_ReplaceKeys'}}) )
		{
			my($work);
#			warn "Can NOT find replacekey in template : $_" unless !defined($pkg->debug('history')) ||
			$pkg->{'_HistoryFile'}{'buffer'} =~ s|$_|{($work,$dirty) = &{$pkg->{'_HistoryFile'}{'_ReplaceKeys'}{$_}}($pkg,"$path/$rh_filename->{'name'}",$ra_buffer,$&);$work;}|sge;
		}
#>>>DEBUG_CODE
		warn $pkg->{'_HistoryFile'}{'buffer'} if defined($pkg->debug('template'));
#>>>DEBUG_CODE
		if( $dirty )# There are some change
		{
			# Write back history
			open(TEMP,"> $tempfile") || die "Can not create temporary file $tempfile: $!";
			print TEMP $pkg->{'_HistoryFile'}{'buffer'};
			close(TEMP);
			rename($tempfile,"$path/$rh_filename->{'name'}");
#>>>DEBUG_CODE
			warn "$path/$rh_filename->{'name'}" if defined($pkg->{'_debug'}{'history'});
#>>>DEBUG_CODE
			push(@updatedfiles,$rh_filename->{'name'});
		}
		# Get next new one
		if( @{$ra_buffer} > 0 )
		{
			$pkg->LoadFile($pkg->{'_TemplateFile'}{'history'},\$pkg->{'_TemplateFile'}{'buffer'}) unless defined($pkg->{'_TemplateFile'}{'buffer'});
			$pkg->{'_HistoryFile'}{'buffer'} = $pkg->{'_TemplateFile'}{'buffer'};
			die "Invalid history filename $rh_filename->{'name'}" unless $rh_filename->{'name'} =~ m/$rh_filename->{'regular_expr'}/;
			$rh_filename->{'name'} = sprintf($rh_filename->{'format'},$1+1);
		}
	}
	return @updatedfiles;
}
#######################
# Field extender 2nd
sub NewItemsExtender
{
my($pkg) = shift;
my($this,$ra_record,$base_string) = @_;
my($lasttime,$count,$format) = ($1,$2,$3);
my($sprintwork);
my(@extended_array);
my(%hash_buffer,$i);
my($replace_match) = join('|',@{$pkg->{'_UpdateInfo'}{'fieldlist'}});
my(@record);
my($append_count) = 0;

	if( @{$ra_record} > 0 )
	{
#>>>DEBUG_CODE
		warn $base_string if defined($pkg->{'_debug'}{'template'});
		warn scalar @{$ra_record} if defined($pkg->{'_debug'}{'records'});
#>>>DEBUG_CODE
		@{$ra_record} = grep { if(m/^(\d+)/s){$1 > $lasttime;}else{0;}} @{$ra_record};
#>>>DEBUG_CODE
		warn $lasttime if defined($pkg->{'_debug'}{'records'});
		warn scalar @{$ra_record} if defined($pkg->{'_debug'}{'records'});
#>>>DEBUG_CODE
		@{$ra_record} = sort {my($a_time,$b_time);$a =~ m/^\d+/; $a_time = $&;$b =~ m/^\d+/; $b_time = $&; $a_time <=> $b_time;} @{$ra_record};

		while( $count < $pkg->{'limit'} && @{$ra_record} > 0 )
		{
			$_ = shift @{$ra_record};
			next if m/^\s*$/;
			++$count;
			++$append_count;

			undef %hash_buffer;
			@record = split(/$pkg->{'_BoundaryString'}{'field'}/,$_);
			$i = 0;
			foreach( @{$pkg->{'_UpdateInfo'}{'fieldlist'}} )
			{
				$hash_buffer{$_} = $record[$i];
				++$i;
			}
			$hash_buffer{'mtime'} = &GetTimeStr($hash_buffer{'mtime'}) if defined($hash_buffer{'mtime'});
			if( defined($pkg->{'base'}) )
			{
				my($uri) = URI->new($hash_buffer{'URL'});
				$hash_buffer{'URL'} = $uri->rel($pkg->{'base'});
			}

			$sprintwork = $format;
			$sprintwork =~ s/\%($replace_match)\%/$hash_buffer{$1}/sgo;
#>>>DEBUG_CODE
			warn $sprintwork if defined($pkg->{'_debug'}{'template'});
#>>>DEBUG_CODE
			unshift(@extended_array,$sprintwork);
		}
		# update information in  base_string 
		if( m/^(\d+)/)
		{
			$lasttime = $1;
			$base_string =~ s/(<!-- NEW items\s+)([0-9]+)(\s+)([0-9]+)(\s+.+-->\s*)/$1$lasttime$3$count$5/s;
#>>>DEBUG_CODE
			warn $base_string if defined($pkg->{'_debug'}{'template'});
#>>>DEBUG_CODE
		}
	}
	unshift(@extended_array,$base_string);
#>>>DEBUG_CODE
	warn $append_count if defined($pkg->{'_debug'}{'records'});
#>>>DEBUG_CODE
	return wantarray ? (join('',@extended_array),$append_count) : join('',@extended_array);
}
#######################
# Field extender 2nd
sub HistoryIndexExtender
{
my($pkg) = shift;
my($this,$ra_record,$base_string) = @_;
my($sprintwork);
my($extended) = '';
my(%hash_buffer,$i);
my($replace_match) = join('|',@{$pkg->{'_HistoryIndex'}{'fieldlist'}},'URL');
my($html_root_match) = join('|',keys(%{$pkg->{'_HtmlRoot'}}));
my(@record);

#>>>DEBUG_CODE
	 if( defined($pkg->{'_debug'}{'template'}) )
	{
		warn scalar @{$ra_record};
		warn $base_string;
	}
#>>>DEBUG_CODE

	@{$ra_record} = reverse sort {my($a_time,$b_time);$a =~ m/^\d+/; $a_time = $&;$b =~ m/^\d+/; $b_time = $&; $a_time <=> $b_time;} @{$ra_record};

	foreach( @{$ra_record} )
	{
		next if m/^\s*$/;
		undef %hash_buffer;
		@record = split(/$pkg->{'_BoundaryString'}{'field'}/,$_);
		$i = 0;
		foreach( @{$pkg->{'_HistoryIndex'}{'fieldlist'}} )
		{
			$hash_buffer{$_} = $record[$i];
			++$i;
		}
		$hash_buffer{'mtime'} = &GetTimeStr($hash_buffer{'mtime'}) if defined($hash_buffer{'mtime'});
		if( defined($hash_buffer{'filename'}) )
		{
			$hash_buffer{'URL'} = &{$pkg->{'_HistoryIndex'}{'File2URL'}}($pkg,$hash_buffer{'filename'});
		}

		$sprintwork = $base_string;
		$sprintwork =~ s/\%($replace_match)\%/$hash_buffer{$1}/sgo;
		$extended .= $sprintwork;
	}
#>>>DEBUG_CODE
	warn $extended if defined($pkg->{'_debug'}{'template'});
#>>>DEBUG_CODE
	return $extended;
}
#######################
# Set URI base
sub SetURIBase
{
my($pkg,$para) = @_;
my($html_root_match) = join('|',keys(%{$pkg->{'_HtmlRoot'}}));
	if( $para  eq  'rel' ) # relative addres mode
	{
		$pkg->{'base'} = $pkg->{'path'}."/";
		$pkg->{'base'} =~ s/^($html_root_match)/$pkg->{'_HtmlRoot'}{$1}/o;
	}
}
#######################
# File path to relative URL
sub File2RelURL
{
my($pkg,$filepath) = @_;
my($html_root_match) = join('|',keys(%{$pkg->{'_HtmlRoot'}}));

	$filepath =~ s/^($html_root_match)/$pkg->{'_HtmlRoot'}{$1}/o;
my($uri) = URI->new($filepath);
	$filepath = $uri->rel($pkg->{'base'});
	return $filepath;
}
#######################
# File path to Absolute URL
sub File2AbsURL
{
my($pkg,$filepath) = @_;
my($html_root_match) = join('|',keys(%{$pkg->{'_HtmlRoot'}}));

	$filepath =~ s/^($html_root_match)/$pkg->{'_HtmlRoot'}{$1}/o;
	return $filepath;
}
1;
